import json
import requests
from pyzenobase import ZenobaseEvent


class ZenobaseAPI:
    # TODO: Keep track of open/closed state internally

    HOST = "https://api.zenobase.com"

    def __init__(self, username=None, password=None):
        payload = {}
        if username is not None:
            payload = {"grant_type": "password", "username": username, "password": password}
        else:
            payload = {"grant_type": "client_credentials"}
        r = requests.post(self.HOST + "/oauth/token", data=payload)
        data = r.json()
        if "error" in data:
            raise Exception("Invalid Zenobase credentials")
        self.access_token = data["access_token"]
        self.client_id = data["client_id"]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _request(self, method, endpoint, data=None, headers={}):
        url = self.HOST + endpoint
        headers["Authorization"] = "Bearer {}".format(self.access_token)
        headers["Content-Type"] = "application/json"
        args = [url]
        kwargs = {"data": json.dumps(data), "headers": headers}
        r = method(*args, **kwargs)
        if not (200 <= r.status_code < 300):
            raise Exception("Status code was not 2xx: {}".format(r))

        if "content-type" in r.headers:
            if "application/json" in r.headers["content-type"]:
                return r.json()
        return r.text

    def _get(self, *args, **kwargs):
        return self._request(requests.get, *args, **kwargs)

    def _post(self, *args, **kwargs):
        return self._request(requests.post, *args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self._request(requests.delete, *args, **kwargs)

    def list_buckets(self, offset=0, limit=100):
        """Limit breaks above 100"""
        # TODO: If limit > 100, do multiple fetches
        if limit > 100:
            raise Exception("Zenobase can't handle limits over 100")
        return self._get("/users/{}/buckets/?order=label&offset={}&limit={}".format(self.client_id, offset, limit))

    def get_bucket(self, bucket_id):
        return self._get("/buckets/{}".format(bucket_id))

    def create_bucket(self, label, description=""):
        if not 1 <= len(label) <= 20:
            raise Exception("Bucket name must be 1-20 chars and can only contain [a-zA-Z0-9-_ ]")
        return self._post("/buckets/", data={"label": label, "description": description})

    def create_or_get_bucket(self, label, description=""):
        data = self.list_buckets()
        for bucket in data["buckets"]:
            if bucket["label"] == label:
                return bucket
        return self.create_bucket(label, description=description)

    def delete_bucket(self, bucket_id):
        self._delete("/buckets/{}".format(bucket_id))

    def list_events(self, bucket_id):
        return self._get("/buckets/{}/".format(bucket_id))

    @staticmethod
    def _bucket_id_from_bucket_or_id(bucket_or_bucket_id):
        assert isinstance(bucket_or_bucket_id, str) or isinstance(bucket_or_bucket_id, dict)
        if isinstance(bucket_or_bucket_id, str):
            bucket_id = bucket_or_bucket_id
        else:  # if isinstance(bucket_or_bucket_id, dict):
            if "@id" in bucket_or_bucket_id:
                bucket_id = bucket_or_bucket_id["@id"]
            else:
                raise Exception("@id field was not in the bucket dict, are you sure you passed a bucket?")
        return bucket_id

    def create_event(self, bucket_or_bucket_id, event):
        assert isinstance(event, ZenobaseEvent) or isinstance(event, dict)
        bucket_id = self._bucket_id_from_bucket_or_id(bucket_or_bucket_id)
        return self._post("/buckets/{}/".format(bucket_id), data=event)

    def create_events(self, bucket_or_bucket_id, events):
        assert isinstance(events, list)
        bucket_id = self._bucket_id_from_bucket_or_id(bucket_or_bucket_id)
        for event in events:
            assert isinstance(event, ZenobaseEvent) or isinstance(event, dict)
        return self._post("/buckets/"+bucket_id+"/", data={"events": events})

    def close(self):
        return self._delete("/authorizations/"+self.access_token)



from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='pyzenobase',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/development.html#single-sourcing-the-version
    version='0.1.0',

    description='A simple API wrapper for Zenobase',
    long_description="For more, see the README.",

    # The project's main homepage.
    url='https://github.com/ErikBjare/pyzenobase',

    # Author details
    author='Erik Bjäreholt',
    author_email='erik@bjareho.lt',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    # What does your project relate to?
    keywords='development zenobase',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=["pyzenobase"],

    # List run-time dependencies here.  These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/technical.html#install-requires-vs-requirements-files
    install_requires=['pytz','requests', 'tzlocal']
)



from datetime import datetime
from pyzenobase import fmt_datetime

_VALID_FIELDS = ["bits", "concentration", "count", "currency", "distance",
                 "distance/volume", "duration", "energy", "frequency", "height",
                 "humidity", "location", "moon", "note", "pace", "percentage",
                 "pressure", "rating", "resource", "sound", "source", "tag",
                 "temperature", "timestamp", "velocity", "volume", "weight"]


class ZenobaseEvent(dict):
    """
        Provides simple structure checking
    """
    def __init__(self, data):
        super(ZenobaseEvent, self).__init__(self)
        for field in data:
            assert field in _VALID_FIELDS
            self[field] = data[field]

        self.clean_data()

    def clean_data(self):
        """Ensures data is Zenobase compatible and patches it if possible,
        if cleaning is not possible it'll raise an appropriate exception"""

        # TODO: Do the same for duration/timedelta
        if "timestamp" in self:
            self._check_timestamp()
            if type(self["timestamp"]) == list:
                def datetime_to_string(dt):
                    return fmt_datetime(dt) if type(dt) == datetime else dt
                self["timestamp"] = list(map(datetime_to_string, self["timestamp"]))
            elif type(self["timestamp"]) == datetime:
                self["timestamp"] = fmt_datetime(self["timestamp"])

        # FIXME: Support list of volumes
        if "volume" in self:
            self["volume"]["unit"] = self["volume"]["unit"].replace("l", "L")

        # FIXME: Support list of weights
        if "weight" in self:
            # Zenobase uses "ug" for micrograms
            if self["weight"]["unit"] in ["mcg", "µg"]:
                self["weight"]["unit"] = "ug"

    def _check_timestamp(self):
        if not type(self["timestamp"]) in (str, datetime, list) and \
                (type(self["timestamp"]) != list or all(map(lambda x: type(x) in (str, datetime), self["timestamp"]))):
            raise TypeError("timestamp must be string, datetime or list of strings/datetimes")




import pytz
from tzlocal import get_localzone


def fmt_datetime(dt, timezone=str(get_localzone())):
    tz = pytz.timezone(timezone)
    dt = dt.astimezone(tz) if dt.tzinfo else tz.localize(dt)
    return dt.strftime('%Y-%m-%dT%H:%M:%S.000%z')
